from Routines import *
from Read_data import *
from pathlib import Path

p=Path(__file__).parent.parent

def CV_divide_vector(num_y_train, size_vali, ncv, basis_size ):
    '''This function takes as input:
    - num_y_train: a numeric vector of y values
    - size_vali: a number, the size of the validation set in each splitting
    - ncv: a number, the number of CV splittings to be generated
    - basis_size: a number, the number of segments on which to base the DG matrix
    this function returns:
    -cv_iterations: a list of tubles ncv long. Each tuple contains a different split.
                    In each tuple the first element is the test array of the splitting,
                     and the second element is the training array of the splitting.'''
    cv_iterations=[]

    for iter in range(ncv):
        miny=np.min(num_y_train)
        maxy=np.max(num_y_train)
        delimiters = np.linspace(miny, maxy, basis_size, endpoint=False)
        digiY = np.digitize(num_y_train, delimiters)
        basis_train_list=[]
        basis_test_list=[]
        for segment in range(1,basis_size+1):
            #In this loop we make sure that each segment is represented in the training set of each cv split
            indices_this_segment=np.where(digiY==segment)[0]
            if len(indices_this_segment)==0:
                print('how is this possible? Reconsider the data and exclude outliers.')
            elif len(indices_this_segment)==1:
                basis_train = indices_this_segment
                basis_train_list.extend(basis_train)
            else:
                basis_train=np.random.choice(indices_this_segment, 2, replace=False)
                basis_train_list.extend(basis_train)
                indices_this_segment_not_tr=[x for x in indices_this_segment if x not in basis_train]
                if len(indices_this_segment_not_tr)>1:
                    basis_test=np.random.choice(indices_this_segment_not_tr,1, replace=False)
                    basis_test_list.extend(basis_test)

        available=[x for x in range(len(num_y_train)) if x not in basis_train_list and x not in basis_test_list]
        vali_add = np.random.choice(available, max(0, size_vali-basis_size), replace = False)
        cvtrain_add = np.array([x for x in available if not x in vali_add])

        vali=np.concatenate((np.array(basis_test_list), np.array(vali_add)))
        cvtrain=np.concatenate((np.array(basis_train_list), np.array(cvtrain_add)))
        cv_iterations.append((vali, cvtrain))

    return cv_iterations



def single_case_regression(tuple_info, residual_df , result_folder, tuber_data_info, ncv, maximum_basis_size, var_field_center_info, singlevar=''):
    '''This function performs regression, it takes as input:
    -tuple_info: a tuple containg two strings, the first regarding the training set (field and year),
                 the second regarding the test set (field and year)
    -residual_df: a dataframe, defined in the main body of the script to store the results of regression
    -result_folder: folder path, where to save plots and dataframes resulting from regression
    -tuber_data_info: a string identifying the type of model to be used in filenames
    -ncv: number of cv splits
    -maximum_basis_size: maximum number of segments for discontinuous Galerkin, if the given number of segments results
                         in a singular matrix the code decreases the number of segments by one until a viable matrix
                         V is obtained
    -var_field_center_info: string, accepted values are "field" or "variety", it refers to how the vitality data is
                            normalized, i.e. weather the y_normalized=y_orig - mean(y_orig) or
                            y_normalized=y_orig - mean( of respective variety)
    -single_var: string, if empty, regression will be performed on the full vector Y, if not empty the only values
                 allowed are full variety names in all caps, then regression will be performed only on data regarding
                 the specified variety.
     and returns residual_df, a dataframe containing the results of regression for each basis size up to maximum,
     if the maximum is 5, there will be 4 rows in the df for the averag of the CV splits, 4 rows for the performance of
     the varying number of segments (2 segments, 3 segments, 4 segments, 5 segments) on the test set, one row for the performance
     of the optimal number of segments on the test set and one row for the model with column selection.'''
    train_info=tuple_info[0]
    test_info=tuple_info[1]

    train_infoy, train_infof, test_on_y, test_on_f, Y, X, Yt, Xt, n_intervals = read_train_test_data(train_info, test_info,maximum_basis_size, var_field_center_info)

    if singlevar!='':
        try:
            train_info = (train_infoy, f'{train_infof}_{singlevar}')
            test_info = (test_on_y, f'{test_on_f}_{singlevar}')
            Y = Y[Y['Variety'] == singlevar]
            X= X[X['Variety'] == singlevar]
            Yt = Yt[Yt['Variety'] == singlevar]
            Xt = Xt[Xt['Variety'] == singlevar]
        except:
            print("The variable singlevar has to contain a valid variety name in allcaps, e.g. 'COLOMBA'")
            sys.exit()
    else:

        train_info=(train_infoy, train_infof)
        test_info=(test_on_y, test_on_f)
    sizedata= Y.shape[0]
    size_vali=int(0.2*sizedata)
    cviter=CV_divide_vector(Y.iloc[:,-1], size_vali, ncv, n_intervals)
    X_columns = X.columns[2:]

    residual_df=experiment_pipeline(Y, Yt, X, Xt, residual_df,cviter, train_info, test_info, tuber_data_info, result_folder, maximum_basis_size, X_columns)

    return residual_df

# the variable var_spec determines whether we are interested in a full field model or variety specific model,
# for the full field model the variable should be equal to the empty string, otherwise it should be the full
# name of the variety of interest in all caps.
var_spec=''
ncv=10
if var_spec:
    var_info='var specific'
    maximum_basis_size = 5
    varname=var_spec
else:
    var_info='var agnostic'
    maximum_basis_size = 7
    varname=''

# the following four variables determine which combination of training and testing data we are interested in
trainy ='2019' # year of training
trainf = 'S'   # field of training
testy = '2020' # year of testing
testf = 'MS'   # field of testing (single fields are M, S, V, the mean of all three is AVG,
               # the pairwise means are MS, SV, MV)



tuber_data_info = f'all - no micro - {maximum_basis_size} segments - field center - {var_info}'
result_folder = os.path.join('./Output', tuber_data_info)

os.makedirs( result_folder,exist_ok=True)

residual_df = pd.DataFrame(columns=['Model', 'Dataset', 'Trained on field', 'Trained on year', 'Tested on field',
                                    'Tested on year','Degree', 'Residual y test', 'Residual X', '# of columns'])


train_case = f'{trainf}_{trainy}'
test_case = f'{testf}_{testy}'

residual_df = single_case_regression((train_case, test_case), residual_df, result_folder, tuber_data_info, ncv,
                                     maximum_basis_size, 'field', varname)
residual_df.to_csv(os.path.join(result_folder, f'{tuber_data_info}.csv'), index=False)

