import numpy as np
import matplotlib.pyplot as plt
import matplotlib
import os


plt.rcParams['font.size'] = '22'
plt.rc('font', **{'family': 'serif', 'serif': ['Computer Modern']})
plt.rc('text', usetex=True)
plt.rcParams['xtick.labelsize'] = 28#30
plt.rcParams['ytick.labelsize'] = 28#30


def plot_col_selection_residuals( glob_colw_res,  filter, optthr,info, result_folder, Xcols):
    '''This function produces plots such as those in figure 5 of the paper, this shows the column-wise residuals
    between the learned representation of the columns of X_test and the original columns as markers, a dashed
    horizontal line crossing the markers represents the optimal threshold, columns whose residual is above the
    line will be discarded.
     '''
    p=filter.shape[0]
    fig, ax=plt.subplots(figsize=(24,14), num=1)

    markersize=5
    colors = matplotlib.colormaps['tab20b'].colors

    datatypes=set([x.split('_')[0] for x in Xcols])
    ci=0
    for datype in datatypes:
        indices_dt=np.where(Xcols.str.startswith(datype))[0]
        ax.semilogy(indices_dt,glob_colw_res[indices_dt], 'o', c=colors[ci],markersize=markersize, label=datype)
        ci+=2
    ax.set_ylabel(r'$\chi_j$')
    ax.set_xlabel(r'Columns of $X$')
    xmin, xmax, ymin, ymax = ax.axis()
    ax.hlines(optthr, xmin=0, xmax=p, linestyle='--',linewidth=3.5, colors='tab:blue', label=r'$\tau_{opt}$' )

    for jj in range(p):
        if ~filter[jj]:
            ax.plot([jj, jj], [ymin, ymax], color='gray', linewidth=0.2, alpha=0.4, zorder=0, label='Discarded')

    handles, labels = ax.get_legend_handles_labels()
    labels=[x.get_label() for x in handles]
    by_label = dict(zip(labels, handles))

    ax.legend(by_label.values(), by_label.keys(),ncol=7,loc='lower center',bbox_to_anchor=(0.5,-0.2),columnspacing=1)# bbox_to_anchor=(1,0.5)

    plt.subplots_adjust(bottom=0.15)
    fig.suptitle(f'{info}')
    try:
        plt.savefig(os.path.join(result_folder, f"{info}.pdf"))
    except:
        print("Pyplot error, couldn't save column selection residuals image")

